@@ -25,6 +25,30 @@ def sin(a):
         raise TypeError
 
 
+def arcsin(a):
+    """
+    Implements sine inverse function to contribute to a computational graph. Converts integers and floats to
+    Node's, and inserts derivatives into the computational graph.
+
+    Args:
+        a (int, float, Node): Value to be calculated through arcsin.
+
+    Returns:
+        Node with values and derivatives along with a corresponding computational graph.
+
+    Raises:
+        TypeError if value is not of type int, float, or Node.
+        """
+    if isinstance(a, Node):
+        return funcNode(np.arcsin, lambda x: 1 / np.sqrt(1 - x * x), None, a, None)
+    elif isinstance(a, int) or isinstance(a, float):
+        n = valNode()
+        n.set_val(np.arcsin(a))
+        return n
+    else:
+        raise TypeError
+
+
 def cos(a):
     """
     Implements cosine to contribute to a computational graph. Converts integers and floats to
@@ -49,6 +73,30 @@ def cos(a):
         raise TypeError
 
 
+def arccos(a):
+    """
+    Implements inverse cosine  function to contribute to a computational graph. Converts integers and floats to
+    Node's, and inserts derivatives into the computational graph.
+
+    Args:
+        a (int, float, Node): Value to be calculated through arccos.
+
+    Returns:
+        Node with values and derivatives along with a corresponding computational graph.
+
+    Raises:
+        TypeError if value is not of type int, float, or Node.
+        """
+    if isinstance(a, Node):
+        return funcNode(np.arccos, lambda x: -1 / np.sqrt(1 - x * x), None, a, None)
+    elif isinstance(a, int) or isinstance(a, float):
+        n = valNode()
+        n.set_val(np.arccos(a))
+        return n
+    else:
+        raise TypeError
+
+
 def tan(a):
     """
     Implements tangent to contribute to a computational graph. Converts integers and floats to
@@ -73,6 +121,30 @@ def tan(a):
         raise TypeError
 
 
+def arctan(a):
+    """
+    Implements inverse tangent  function to contribute to a computational graph. Converts integers and floats to
+    Node's, and inserts derivatives into the computational graph.
+
+    Args:
+        a (int, float, Node): Value to be calculated through arctan.
+
+    Returns:
+        Node with values and derivatives along with a corresponding computational graph.
+
+    Raises:
+        TypeError if value is not of type int, float, or Node.
+        """
+    if isinstance(a, Node):
+        return funcNode(np.arctan, lambda x: 1 / (1 + x * x), None, a, None)
+    elif isinstance(a, int) or isinstance(a, float):
+        n = valNode()
+        n.set_val(np.arctan(a))
+        return n
+    else:
+        raise TypeError
+
+
 def sinh(a):
     """
     Implements hyperbolic sine to contribute to a computational graph. Converts integers and floats to
@@ -175,7 +247,7 @@ def logistic(a):
         Node's, and inserts derivatives into the computational graph.
 
         Args:
-            a (int, float, Node): Value to be calculated through tan.
+            a (int, float, Node): Value of which we are calulating logistic.
 
         Returns:
             Node with values and derivatives along with a corresponding computational graph.
@@ -197,7 +269,7 @@ def _logistic(a):
     """
     Helper function for logistic, implements logistic function for int or float data type.
     Args:
-        a (int, float, Node): Real number whose logistic function is to be calculated.
+        a (int, float): Real number whose logistic function is to be calculated.
 
     Returns:
         Logistic value of the argument a.
@@ -265,6 +337,16 @@ def ln(a):
 
 
 def _log_ab(arg, base=np.exp(1)):
+    """
+        Helper function for log_ab, implements logarithm function for int or float data type with arg as argument
+        of logarithm and base as the base.
+        Args:
+            arg (int, float): Argument of the logarithm function.
+            base (int, float): Base of the logarithm function.
+        Returns:
+            log value of the argument arg and base base .
+    """
+
     return np.log(arg) / np.log(base)
 
 
@@ -293,19 +375,19 @@ def log_ab(a, b):
         n = valNode()
         n._set_val((np.log(a)) / np.log(b))
         return n
+    else:
+        raise TypeError
 
 
 class Node:
     """
     Node class that implements the following native python functions: addition, subtraction, multipliation,
     division, unary operators, and power. Depending on the implemented funtions, Node will check if the inputs
-    are of type integer, float, or Node. If they are of tyep Node, then the value and it's derivative is computed
-    and stored in a dictionary. If the values are of type int or float, then the values are first convered to Node
-    class by calling valNode.
+    are of type integer, float, or Node. If they are of tyep Node, then a new Node will be generated to be the
+    parent of inputs and be returned. If the values are of type int or float, then the values are first convered
+    to Node class by calling valNode, and be operated in the same way as when the inputs are of type Node.
 
     Attributes:
-        val (int, float): Current value of the variable.
-        der (int, float): Derivative of variable.
         add: Addition implementation for Node class.
         sub: Subtraction implementation for Node class.
         mul: Multiplication implementation for Node class.
@@ -318,7 +400,6 @@ class Node:
     """
 
     def __init__(self):
-        # self.der = dict()
         pass
 
     def __add__(self, other):
@@ -454,16 +535,18 @@ class Node:
 class valNode(Node):
     """
     Sets variables as valNode class, which allows the users to auto-differentiate functions that use them. valNode class
-    overrides native python functions and unary operators, in order to construct a sequential computational graph. Sets
-    the initial derivative (0) for each assigned variable.
+    overrides native python functions and unary operators, in order to construct a sequential computational graph.
+
+    Attributes:
+        val (int, float): Current value of the variable.
+        name (str): The name of the variable.
+        der (int, float): Derivative of variable, which is used in the reverse mode.
     """
 
     def __init__(self, name=None):
         super().__init__()
         self.der = 0
         self.name = name
-        # if name != None:
-        #     self.der[name]=1
 
     def _set_val(self, val):
         """
@@ -482,7 +565,10 @@ class valNode(Node):
             val(int, float): numeric value of the valNode.
 
         """
-        self._set_val(val)
+        if isinstance(val, int) or isinstance(val, float):
+            self._set_val(val)
+        else:
+            raise TypeError("The value for a variable should be 'int' or 'float'.")
 
     def __str__(self):
         """
@@ -498,27 +584,47 @@ class valNode(Node):
         Executes a simple forward pass for a single node.
 
         Returns:
-            Current value and empty derivative if number. Otherwise returns the value of the specified valNode and
-            empty dictionary.
+            Current value and empty derivative if no name. Otherwise returns the value of the specified valNode and
+            derivative 1.
         """
         if self.name is None:
-            return self.val, {self.name: 0}
+            return self.val, {}
         return self.val, {self.name: 1}
 
     def forward_pass(self):
         self.der = 0
         return self.val
-        
+
     def reverse_pass(self, partial, adjoint):
         if self.name == None:
             return dict()
-        self.der += partial*adjoint
+        self.der += partial * adjoint
         return {self.name: self}
 
     def _reset_der(self):
         self.der = 0
 
+
 class funcNode(Node):
+    """
+    Sets intermediate variables as funcNode class to memorize the functions needed to represent the final output 
+    function. funcNode class overrides native python functions and unary operators, in order to construct a 
+    sequential computational graph, based on which the users can auto-differentiate the functions they are interested
+    in.
+
+    Attributes:
+        val (int, float): Current value of the intermediate variable.
+        func (callable): The arithmetic function executed to calculate the intermediate variable.
+        leftdf (callable): The arithmetic function to calculate the partial derivative of the current node 
+        with respect to its left child.
+        rightdf (callable): The arithmetic function to calculate the partial derivative of the current node 
+        with respect to its right child.
+        left (Node): Left child of the node.
+        right (Node): Right child of the node.
+
+
+    """
+
     def __init__(self, func, leftdf, rightdf, left, right):
         super().__init__()
         self.func = func
@@ -534,6 +640,17 @@ class funcNode(Node):
                '\n|\n|-(R)->' + '\n|      '.join(str(self.right).split('\n'))
 
     def forward(self):
+        """
+            calculate function value and derivatives with forward mode.
+            Args:
+                self (Node): The current node.
+
+            Returns:
+                val (float): The value of the function
+                der (dict): The dictionary containing all the partial derivatives of the functions with 
+                variable names as keys.
+
+        """    
         if self.right != None:
             aval, ader = self.left.forward()
             bval, bder = self.right.forward()
@@ -561,212 +678,149 @@ class funcNode(Node):
         else:
             self.val = self.func(self.left.forward_pass())
         return self.val
-    
+
     def reverse_pass(self, partial, adjoint):
-        if self.right != None: 
+        if self.right != None:
             lder = self.leftdf(self.left.val, self.right.val)
             rder = self.rightdf(self.left.val, self.right.val)
-            lvars = self.left.reverse_pass(lder, partial*adjoint)
-            rvars = self.right.reverse_pass(rder, partial*adjoint)
-            variables = dict()
+            lvars = self.left.reverse_pass(lder, partial * adjoint)
+            rvars = self.right.reverse_pass(rder, partial * adjoint)
             for v in rvars:
                 if v not in lvars:
                     lvars[v] = rvars[v]
         else:
             lder = self.leftdf(self.left.val)
-            lvars = self.left.reverse_pass(lder, partial*adjoint) 
-            
+            lvars = self.left.reverse_pass(lder, partial * adjoint)
+
         return lvars
-        # return self.val
-    
+
     def reverse(self):
+        """
+            calculate function value and derivatives with reverse mode.
+            Args:
+                self (Node): The current node.
+
+            Returns:
+                val (float): The value of the function
+                der (dict): The dictionary containing all the partial derivatives of the functions with 
+                variable names as keys.
+
+        """    
         self.forward_pass()
-        variables = self.reverse_pass(1,1)
-        der = { v: node.der for v, node in variables.items() }
+        variables = self.reverse_pass(1, 1)
+        der = {v: node.der for v, node in variables.items()}
         return self.val, der
-    
-    
+
+
 class vector:
+    """
+        Combines several valNodes together to support vectorized input variables. vector and valNode share
+        the same syntaxes for .set_val, .der and ._reset_der. Every single element in the vector variable 
+        can be indexed.
+    """
+    
     def __init__(self, *args):
         self.size = len(args)
         self.elements = args
-    
+
     def set_val(self, array):
+        """
+            set the value of the vector variable.
+            Args:
+                array (list): A list of value to be set for the vector.
+
+            Returns:
+                None
+
+            Raises:
+                ValueError if the input size does not match the vector size.
+        """
         if len(array) != self.size:
             raise ValueError(f"Input size has a mismatch with the vector size ({self.size})")
         for node, val in zip(self.elements, array):
-            # print(val)
             node.set_val(val)
 
     def __getitem__(self, key):
         return self.elements[key]
-        
+
     def __iter__(self, key):
         return iter(self.elements)
-    
+
     def grad(self, var):
-        '''
+        """
             auto-diff for vector functions
             currently we only use reverse mode to calculate gradient for vector functions
-            input: the variable or vector variables var, which is used to
-             define the function and for which we need to calculate the derivatives
-            return: Jacobian matrix
-        '''
+            Args: 
+                the variable or vector variables var, which is used to define the function 
+                and for which we need to calculate the derivatives
+            Return: Jacobian matrix
+        """
         der_array = []
         var._reset_der()
         for f in self.elements:
             f.forward_pass()
-            f.reverse_pass(1,1)
+            f.reverse_pass(1, 1)
             der_array.append(var.der)
             var._reset_der()
         return der_array
-    
+
     def evaluate(self):
-        '''
+        """
             evaluation for vector functions
-        '''
-        return [ f.forward_pass() for f in self.elements ]        
-    
+            Return: The vector values of the vector function.
+        """
+        return [f.forward_pass() for f in self.elements]
+
     @property
     def der(self):
-        return [ node.der for node in self.elements ]
-    
+        return [node.der for node in self.elements]
+
     def _reset_der(self):
-        '''
+        """
            this function only works if it is a variable vector
-        '''
+        """
         for node in self.elements:
             node.der = 0
-    
-def variables(name, size = None):
-    if size == None or size == 1:
+
+
+def variables(name, size=None):
+    """
+        initialize independent variables for the function of interest
+        
+        Args:
+            name (str): The name to identify the variable.
+            size (int): The size of the variable.
+
+        Returns:
+            vector when size > 1
+            valNode when size = 1 or None
+    """
+    if (size is None) or (size == 1):
         return valNode(name)
-    elems = [ valNode(name+'___'+str(i)) for i in range(size) ]
+    elems = [valNode(name + '___' + str(i)) for i in range(size)]
     return vector(*elems)
 
-    
-
-            
 
 if __name__ == '__main__':
-    # v = vector([1,2,3])
+    def func(v):    
+        # f takes a size=3 vector and output a size=2 vector
+        f1 = sin(ln(v[0])) + tan(v[0] ** 2 + v[0] * v[1] + v[0] ** 3 * v[1])
+        f2 = v[0] * v[1] + v[1] ** 2 + sqrt(v[0])
+        return vector(f1, f2)
 
-    # example to define scalar function with multiple input variables
 
-    x = variables('x')
-    y = variables('y')
-    f = sin(ln(x))+tan(x*x+y*x+x*x*x*y)
-    x.set_val(2)
-    y.set_val(3)
-    print(f.forward())
-    print(f.reverse())
-    
-    # example to define vector function with vectorized input variables
-    
-    v = variables('v', 2)
-    v.set_val([2,3])    
-    print(v.evaluate())
-    print(v.grad(v))
-    def func(v):
-        '''
-            f takes a size=3 vector and output a size=2 vector
-        '''
-        f1 = sin(ln(v[0]))+tan(v[0]**2+v[0]*v[1]+v[0]**3*v[1])
-        f2 = v[0]*v[1]+v[1]**2+sqrt(v[0])
-        return vector(f1,f2)
     f = func(v)
     print(f.evaluate())
     print(f.grad(v))
-        
-        
+
+
     # vector function can also take a scalar input variable
     def func1(x):
-        f1 = x**2
-        f2 = x**3
+        f1 = x ** 2
+        f2 = x ** 3
         return vector(f1, f2)
-    
+
+
     f = func1(x)
     print(f.evaluate())
     print(f.grad(x))
-        
-    '''x.der = 0
-    y.der = 0
-    f = log(sin(exp(x)+y)) + exp(y) + x
-    x._set_val(0)
-    y._set_val(1)
-    print(f.forward())
-    f.forward_pass()
-    f.reverse(1,1)
-    print(f.val, x.der, y.der)
-    
-    x.der = 0
-    y.der = 0
-    f = log(sin(x+y)**2)+exp(x**2+y**2)
-    print(f)
-    x._set_val(1)
-    y._set_val(2)
-    print(f.forward())
-    f.forward_pass()
-    f.reverse(1,1)
-    print(f.val, x.der, y.der)
-    
-    print("testing: sin(ab + b)")
-    a = valNode('a')
-    b = valNode('b')
-    f = sin(a * b + b)
-    a._set_val(2)
-    b._set_val(5)
-    # print(f)
-    # print(f.forward())
-    f_val, f_grad = f.forward()
-
-    actual_f_val = np.sin(a.val * b.val + b.val)
-    actual_f_grad = {
-        'a': b.val * np.cos((a.val + 1) * b.val),
-        'b': (a.val + 1) * np.cos((a.val + 1) * b.val)
-    }
-
-    assert np.isclose(f_val, actual_f_val)
-
-    for var in f_grad:
-        assert np.isclose(f_grad[var], actual_f_grad[var])
-
-    f.forward_pass()
-    # print(f.val, )
-    f.reverse(1, 1)
-    reverse_grads = {
-        'a': a.der,
-        'b': b.der
-    }
-    for var in f_grad:
-        assert np.isclose(f_grad[var], reverse_grads[var])
-
-    print("testing: e^(a/c) + b")
-    a = valNode('a')
-    b = valNode('b')
-    c = valNode('c')
-    f = exp(a / c) + b
-    a._set_val(np.pi / 2)
-    b._set_val(np.pi / 3)
-    c._set_val(np.pi)
-
-    actual_f_val = np.exp(a.val / c.val) + b.val
-    # print(np.exp(a.val / c.val), np.arccos(np.exp(a.val / c.val)))
-
-    actual_f_grad = {
-        'a': (np.exp(a.val / c.val)) / (c.val),
-        'b': 1,
-        'c': -(a.val * np.exp(a.val / c.val)) / (c.val ** 2),
-    }
-
-    f.forward_pass()
-    # print(f.val, )
-    f.reverse(1, 1)
-    reverse_grads = {
-        'a': a.der,
-        'b': b.der,
-        'c': c.der,
-    }
-
-    for var in actual_f_grad.keys():
-        assert np.isclose(actual_f_grad[var], reverse_grads[var])'''
